import subuserlib.classes.user
import subuserlib.classes.gitRepository
import subuserlib.classes.fileStructure
import os
u = subuserlib.classes.user.User()
g = subuserlib.classes.gitRepository.GitRepository(u,os.path.expanduser("~/.subuser/repositories/default/"))
fs = subuserlib.classes.fileStructure.BasicFileStructure(os.path.expanduser("~/.subuser/repositories/default"))
fsac = g.getFileStructureAtCommit("latest")



#!/usr/bin/python3
# -*- coding: utf-8 -*-

#external imports
#import ...
#internal imports
import subuserlib.builtInCommands.list

@subuserlib.profile.do_cprofile
def runCommand(sysargs):
  subuserlib.builtInCommands.list.runCommand(sysargs)



# -*- coding: utf-8 -*-

"""
A subuser is an entity that runs within a Docker container and has a home directory and a set of permissions that allow it to access a limited part of the host system.
"""

#external imports
import os
import stat
import errno
import json
import sys
import collections
#internal imports
import subuserlib.permissions
from subuserlib.classes.userOwnedObject import UserOwnedObject
from subuserlib.classes.permissions import Permissions
from subuserlib.classes.describable import Describable
from subuserlib.classes.subuserSubmodules.run.runtime import Runtime
from subuserlib.classes.subuserSubmodules.run.x11Bridge import X11Bridge
from subuserlib.classes.subuserSubmodules.run.runReadyImage import RunReadyImage
from subuserlib.classes.subuserSubmodules.run.runtimeCache import RuntimeCache

class Subuser(UserOwnedObject, Describable):
  def __init__(self,user,name,imageId,executableShortcutInstalled,locked,serviceSubuserNames,imageSource=None,imageSourceName=None,repoName=None,entrypointsExposed=False,nonDefaultHomeDir=None):
    self.name = name
    self.__imageSource = imageSource
    self.__repoName = repoName
    self.__imageSourceName = imageSourceName
    self.imageId = imageId
    self.executableShortcutInstalled = executableShortcutInstalled
    self.__entryPointsExposed = entrypointsExposed
    self.__entryPointsExposedThisRun = False
    self.locked = locked
    self.serviceSubuserNames = serviceSubuserNames
    self.__x11Bridge = None
    self.__runReadyImage = None
    self.__runtime = None
    self.__runtimeCache = None
    self.__permissions = None
    self.__permissionsTemplate = None
    self.nonDefaultHomeDir = nonDefaultHomeDir
    UserOwnedObject.__init__(self,user)

  @property
  def imageSource(self):
    """
    Note, it is posssible that the subuser's image source no longer exists, in which case this function raises a NoImageSourceException.
    """
    if self.__imageSource is None:
      try:
        self.__imageSource = self.user.registry.repositories[self.__repoName][self.__imageSourceName]
      except KeyError:
        raise NoImageSourceException()
    return self.__imageSource

  @imageSource.setter
  def imageSource(self,new_image_source):
    self.__imageSource=new_image_source
    self.imageId = None

  @property
  def imageSourceName(self):
    if self.__imageSource is None:
      return self.__imageSourceName
    else:
      return self.imageSource.name

  @property
  def sourceRepoName(self):
    if self.__imageSource is None:
      return self.__repoName
    else:
      return self.imageSource.repo.name

  @property
  def entryPointsExposed(self):
    return self.__entryPointsExposed

  @entryPointsExposed.setter
  def entrypointsExposed(self,exposed):
    self.__entryPointsExposed = exposed
    if exposed:
      self.__entryPointsExposedThisRun = True

  def wereEntryPointsExposedThisRun(self):
    return self.__entryPointsExposedThisRun

  @property
  def permissionsDir(self):
    return os.path.join(self.user.config["registry-dir"],"permissions",self.name)

  @property
  def relativePermissionsDir(self):
    """
    Get the permissions directory as relative to the registry's git repository.
    """
    return os.path.join("permissions",self.name)

  def createPermissions(self,permissionsDict):
    permissionsDotJsonWritePath = os.path.join(self.permissionsDir,"permissions.json")
    self.__permissions = Permissions(self.user,initialPermissions=permissionsDict,writePath=permissionsDotJsonWritePath)
    return self.__permissions

  @property
  def permissionsDotJsonWritePath(self):
    return os.path.join(self.permissionsDir,"permissions.json")

  def loadPermissions(self):
    registryFileStructure = self.user.registry.gitRepository.getFileStructureAtCommit(self.user.registry.gitReadHash)
    try:
      initialPermissions = subuserlib.permissions.load(permissionsString=registryFileStructure.read(os.path.join(self.relativePermissionsDir,"permissions.json")),logger=self.user.registry)
    except OSError:
      raise SubuserHasNoPermissionsException("The subuser <"+self.name+"""> has no permissions.

Please run:

$ subuser repair

To repair your subuser installation.\n""")
    except SyntaxError as e:
      sys.exit("The subuser <"+self.name+""">'s permissions appear to be corrupt.

Please file a bug report explaining how you got here.\n"""+ str(e))
    self.__permissions = Permissions(self.user,initialPermissions,writePath=self.permissionsDotJsonWritePath)

  @property
  def permissions(self):
    if self.__permissions is None:
      self.loadPermissions()
    return self.__permissions

  def getPermissionsTemplate(self):
    if self.__permissionsTemplate is None:
      permissionsDotJsonWritePath = os.path.join(self.permissionsDir,"permissions-template.json")
      registryFileStructure = self.user.registry.gitRepository.getFileStructureAtCommit(self.user.registry.gitReadHash)
      if "permissions-template.json" in registryFileStructure.lsFiles(self.relativePermissionsDir):
        initialPermissions = subuserlib.permissions.load(permissionsString=registryFileStructure.read(os.path.join(self.relativePermissionsDir,"permissions-template.json")),logger=self.user.registry)
        save = False
      else:
        initialPermissions = self.imageSource.permissions
        save = True
      self.__permissionsTemplate = Permissions(self.user,initialPermissions,writePath=permissionsDotJsonWritePath)
      if save:
        self.__permissionsTemplate.save()
    return self.__permissionsTemplate

  def editPermissionsCLI(self):
    while True:
      self.user.endUser.runEditor(self.permissions.writePath)
      try:
        initialPermissions = subuserlib.permissions.load(permissionsFilePath=self.permissionsDotJsonWritePath)
        break
      except SyntaxError as e:
        print(e)
        input("Press ENTER to edit the permission file again.")
    self.__permissions = Permissions(self.user,initialPermissions,writePath=self.permissionsDotJsonWritePath)
    self.permissions.save()

  def removePermissions(self):
    """
    Remove the user set and template permission files.
    """
    try:
      self.user.registry.gitRepository.run(["rm",os.path.join(self.relativePermissionsDir,"permissions.json"),os.path.join(self.relativePermissionsDir,"permissions-template.json")])
    except subuserlib.classes.gitRepository.GitException:
      pass
    try:
      self.user.registry.gitRepository.run(["rm",os.path.join(self.relativePermissionsDir,"permissions.json"),os.path.join(self.relativePermissionsDir,"permissions-template.json")])
    except subuserlib.classes.gitRepository.GitException:
      pass

  def isImageInstalled(self):
    return self.user.dockerDaemon.getImageProperties(self.imageId) is not None

  def getRunReadyImage(self):
    if not self.__runReadyImage:
      self.__runReadyImage = RunReadyImage(self.user,self)
    return self.__runReadyImage

  @property
  def x11Bridge(self):
    """
    Return the X11 bridge object for this subuser.
    """
    if not self.__x11Bridge:
      self.__x11Bridge = X11Bridge(self.user,self)
    return self.__x11Bridge

  #TODO clean this up so that we aren't forwarding arguments in needless fluff.
  def getRuntime(self,environment,extraDockerFlags=None,entrypoint=None):
    """
    Returns the subuser's Runtime object for it's current permissions, creating it if necessary.
    """
    if not self.__runtime:
      self.__runtime = Runtime(self.user,subuser=self,environment=environment,extraDockerFlags=extraDockerFlags,entrypoint=entrypoint)
    return self.__runtime

  def getRuntimeCache(self):
    if not self.__runtimeCache:
      self.__runtimeCache = RuntimeCache(self.user,self)
    else:
      self.__runtimeCache.reload()
    return self.__runtimeCache

  def setupHomeDir(self):
    """
    Sets up the subuser's home dir, along with creating symlinks to shared user dirs.
    """
    if self.permissions["basic-common-permissions"] and self.permissions["basic-common-permissions"]["stateful-home"]:
      try:
        self.user.endUser.makedirs(self.homeDirOnHost)
      except OSError as e:
        if e.errno() == errno.EEXIST:
          pass
      if self.permissions["user-dirs"]:
        for userDir in self.permissions["user-dirs"]:
          symlinkPath = os.path.join(self.homeDirOnHost,userDir)
          # https://stackoverflow.com/questions/15718006/check-if-directory-is-symlink
          if symlinkPath.endswith("/"):
            symlinkPath = symlinkPath[:-1]
          destinationPath = os.path.join("/subuser/userdirs",userDir)
          if not os.path.islink(symlinkPath):
            if os.path.exists(symlinkPath):
              self.user.endUser.makedirs(os.path.join(self.homeDirOnHost,"subuser-user-dirs-backups"))
              os.rename(symlinkPath,os.path.join(self.homeDirOnHost,"subuser-user-dirs-backups",userDir))
            try:
              os.symlink(destinationPath,symlinkPath)
              # Arg, why are source and destination switched?
              # os.symlink(where does the symlink point to, where is the symlink)
              # I guess it's to be like cp...
            except OSError:
              pass

  @property
  def homeDirOnHost(self):
    """
    Returns the path to the subuser's home dir. Unless the subuser is configured to have a stateless home, in which case returns None.
    """
    if self.permissions["basic-common-permissions"]["stateful-home"]:
      if self.nonDefaultHomeDir is not None:
        return self.nonDefaultHomeDir
      return os.path.join(self.user.config["subuser-home-dirs-dir"],self.name)
    else:
      return None

  @property
  def dockersideHome(self):
    if self.permissions["as-root"]:
      return "/root/"
    else:
      return "/home/subuser/"

  def describe(self):
    print("Subuser: "+self.name)
    print("------------------")
    try:
      print(self.imageSource.getIdentifier())
    except subuserlib.classes.subuser.NoImageSourceException:
      print("Warning: This subuser has no image, nor does it have a valid image source to install an image from.")
    print("Docker image Id: "+str(self.imageId))
    self.permissions.describe()
    print("")

  def installLaunchScript(self,name,script):
    """
    Install a trivial executable script into the PATH which launches the subser image.
    """
    redirect = script
    executablePath=os.path.join(self.user.config["bin-dir"], name)
    with self.user.endUser.get_file(executablePath, 'w') as file_f:
      file_f.write(redirect)
      st = os.stat(executablePath)
      os.chmod(executablePath, stat.S_IMODE(st.st_mode) | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

  def installExecutableShortcut(self):
    self.installLaunchScript(self.name,"#!/bin/bash\nsubuser run "+self.name+" $@")

  def exposeEntrypoints(self):
    """
    Create launcher executables for the subuser's entrypoints.
    """
    for name,path in self.permissions["entrypoints"].items():
      launchScript = """#!/usr/bin/python3
import subprocess,sys
run = """+json.dumps({"subuser-name":self.name,"entrypoint":path})
      launchScript +="""
sys.exit(subprocess.call(["subuser","run","--entrypoint="+run["entrypoint"],run["subuser-name"]]))
"""
      self.installLaunchScript(name,launchScript)

class SubuserHasNoPermissionsException(Exception):
  pass

class NoImageSourceException(Exception):
  pass



# -* coding: utf-8 -*-

"""
This is a Class which allows one to manipulate a git repository.
"""

#external imports
import os
import tempfile
import sys
import errno
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
from subuserlib.classes.fileStructure import FileStructure
import subuserlib.executablePath
import subuserlib.test
if subuserlib.test.testing:
  hashtestDir = subuserlib.test.hashtestDir
  def getUser():
    return subuserlib.test.getUser()

gitExecutable = None

class GitRepository(UserOwnedObject):
  def __init__(self,user,path):
    UserOwnedObject.__init__(self,user)
    self.path = path
    self.__gitExecutable = None

  def getGitExecutable(self):
    global gitExecutable
    if gitExecutable is None:
      gitExecutable = [subuserlib.executablePath.which("git",excludeDir=self.user.config["bin-dir"])]
      if gitExecutable is None:
        sys.exit("Git is not installed. Subuser requires git to run.")
    return gitExecutable

  def assertGitSetup(self):
    self.getGitExecutable()
    def getConfig(key):
      (returncode,stdout,stderr) = self.user.endUser.callCollectOutput(gitExecutable + ["config","--get",key])
      if returncode == 0:
        return stdout
      else:
        return None
    if getConfig("user.name") is None or getConfig("user.email") is None:
      sys.exit("""You must configure git's user.name and user.email options before using subuser.
Do so by running
$ git config --global user.name "John Doe"
$ git config --global user.email johndoe@example.com
""")

  def clone(self,origin):
    """
    Clone an external repository in order to create this repository.
    """
    return self.user.endUser.call(self.getGitExecutable()+["clone", origin, self.path])

  def run(self,args):
    """
    Run git with the given command line arguments.
    """
    try:
      gitArgs = self.getGitExecutable()+args
      (returncode,stdout,stderr) = self.user.endUser.callCollectOutput(gitArgs,cwd=self.path)
      self.user.registry.log(self.path+": "+" ".join(gitArgs),verbosityLevel=3)
      self.user.registry.log(stdout,verbosityLevel=3)
      self.user.registry.log(stderr,verbosityLevel=3)
      if stderr:
        raise GitException(stderr)
      return returncode
    except OSError as e:
      if e.errno == errno.EEXIST:
        sys.exit("You must have git installed to use subuser.")
      else:
        raise e

  def runShowOutput(self,args):
    self.user.endUser.call(self.getGitExecutable()+args,cwd=self.path)

  def runCollectOutput(self,args,eatStderr=False):
    """
    Run git with the given command line arguments and return a tuple with (returncode,output).
    """
    try:
      gitArgs = self.getGitExecutable()+args
      (returncode,stdout,stderr) = self.user.endUser.callCollectOutput(gitArgs,cwd=self.path)
      self.user.registry.log(self.path+": "+" ".join(gitArgs),verbosityLevel=3)
      self.user.registry.log(stderr,verbosityLevel=3)
      if stderr and not eatStderr:
        raise GitException(stderr)
      return (returncode,stdout)
    except OSError as e:
      if e.errno == errno.EEXIST:
        sys.exit("You must have git installed to use subuser.")
      else:
        raise e

  def doesCommitExist(self,commit):
    """
    Return true if the commit or reference exists.
    """
    try:
      (returncode,stdout) = self.runCollectOutput(["cat-file","-t",commit])
    except GitException:
      return False
    if returncode == 0 and "commit" in stdout:
      return True
    else:
      return False

  def getFileStructureAtCommit(self,commit):
    """
    Get a ``FileStructure`` object which relates to the given git commit.
    """
    return GitFileStructure(self,commit)

  def commit(self,message):
    """
    Run git commit with the given commit message.
    """
    return self.run(["commit","--message",message])

  def checkout(self,commit,files=[]):
    """
    Run git checkout
    """
    self.run(["checkout",commit]+files)

  def getHashOfRef(self,ref):
    command = ["show-ref","-s",ref]
    (returncode,output) = self.runCollectOutput(command)
    if returncode != 0:
      raise OSError("Running git in "+self.path+" with args "+str(command)+" failed.")
    return output.strip()

class GitFileStructure(FileStructure):
  def __init__(self,gitRepository,commit):
    """
    Initialize the file structure.

    Here we setup test stuff:
    >>> import subuserlib.subprocessExtras
    >>> import subuserlib.classes.gitRepository
    >>> subuserlib.subprocessExtras.call(["git","init"],cwd=subuserlib.classes.gitRepository.hashtestDir)
    0
    >>> subuserlib.subprocessExtras.call(["git","add","."],cwd=subuserlib.classes.gitRepository.hashtestDir)
    0
    >>> subuserlib.subprocessExtras.call(["git","commit","-m","Initial commit"],cwd=subuserlib.classes.gitRepository.hashtestDir)
    0
    """
    self.gitRepository = gitRepository
    self.commit = commit
    self.__lsTreeCache = {}

  def lsTree(self):
    """
    Returns a list of tuples of the form:
    (mode,type,hash,path)

    Coresponding to the items found in the subfolder.
    """
    args = [self.commit,"-rtl"]
    argsTuple = tuple(args)
    try:
      return self.__lsTreeCache[argsTuple]
    except KeyError:
      pass
    (returncode,output) = self.gitRepository.runCollectOutput(["ls-tree"]+args)
    if returncode != 0:
      return [] # It is simpler to just return [] here than to check if the repository is properly initialized everywhere else.
    lines = output.splitlines()
    items = []
    for line in lines:
      mode,objectType,objectHash,size,path = line.split(maxsplit=4)
      line = {"mode":mode,"type":objectType,"hash":objectHash,"size":size,"path":path}
      items.append(line)
    self.__lsTreeCache[argsTuple] = items
    return items

  def _ls(self, subfolder,objectType=None):
    """
    Returns a list of file and folder names.

    >>> from subuserlib.classes.gitRepository import GitRepository
    >>> gitRepository = GitRepository(subuserlib.classes.gitRepository.getUser(),subuserlib.classes.gitRepository.hashtestDir)
    >>> fileStructure = gitRepository.getFileStructureAtCommit("master")
    >>> print(",".join(fileStructure.ls("./")))
    bar,blah
    """
    if subfolder == "./" or subfolder == "/":
      subfolder = ""
    items = self.lsTree()
    names = []
    for item in items:
      folder,name = os.path.split(item["path"])
      if os.path.normpath(folder) == os.path.normpath(subfolder):
        if objectType is None or objectType == item["type"]:
          names.append(name)
    return names

  def _lsFiles(self,subfolder):
    """
    Returns a list of paths to files in the subfolder.
    Paths are relative to the repository as a whole.

    >>> from subuserlib.classes.gitRepository import GitRepository
    >>> gitRepository = GitRepository(subuserlib.classes.gitRepository.getUser(),subuserlib.classes.gitRepository.hashtestDir)
    >>> fileStructure = gitRepository.getFileStructureAtCommit("master")
    >>> print(",".join(fileStructure.lsFiles("./")))
    blah
    """
    return self.ls(subfolder,"blob")

  def _lsFolders(self,subfolder):
    """
    Returns a list of paths to folders in the subfolder.
    Paths are relative to the repository as a whole.

    >>> from subuserlib.classes.gitRepository import GitRepository
    >>> gitRepository = GitRepository(subuserlib.classes.gitRepository.getUser(),subuserlib.classes.gitRepository.hashtestDir)
    >>> fileStructure = gitRepository.getFileStructureAtCommit("master")
    >>> print(",".join(fileStructure.lsFolders("./")))
    bar
    """
    return self.ls(subfolder,"tree")

  def _exists(self,path):
    """
    >>> from subuserlib.classes.gitRepository import GitRepository
    >>> gitRepository = GitRepository(subuserlib.classes.gitRepository.getUser(),subuserlib.classes.gitRepository.hashtestDir)
    >>> fileStructure = gitRepository.getFileStructureAtCommit("master")
    >>> fileStructure.exists("./blah")
    True
    >>> fileStructure.exists("./non-existant")
    False
    """
    (dir,filename) = os.path.split(path)
    for object in self.lsTree():
      (_,existing_filename) = os.path.split(object["path"])
      if filename == existing_filename:
        return True
    return False

  def _read(self,path):
    """
    Returns the contents of the given file at the given commit.

    >>> from subuserlib.classes.gitRepository import GitRepository
    >>> gitRepository = GitRepository(subuserlib.classes.gitRepository.getUser(),subuserlib.classes.gitRepository.hashtestDir)
    >>> fileStructure = gitRepository.getFileStructureAtCommit("master")
    >>> print(fileStructure.read("./blah"))
    blahblah
    <BLANKLINE>
    """
    (errorcode,content) = self.gitRepository.runCollectOutput(["show",self.commit+":"+path],eatStderr=True)
    if errorcode != 0:
      raise OSError("Git show exited with error "+str(errorcode)+". File does not exist.\nRepo:"+self.gitRepository.path+"\nPath: "+path+"\nCommit: "+self.commit+"\n")
    return content

  def _readBinary(self,path):
    return self.read(path).encode()

  def _getMode(self,path):
    """
    >>> from subuserlib.classes.gitRepository import GitRepository
    >>> gitRepository = GitRepository(subuserlib.classes.gitRepository.getUser(),subuserlib.classes.gitRepository.hashtestDir)
    >>> fileStructure = gitRepository.getFileStructureAtCommit("master")
    >>> print(fileStructure.getModeString("./blah"))
    100644
    """
    allObjects = self.lsTree()
    for treeObject in allObjects:
      if os.path.normpath(treeObject["path"]) == os.path.normpath(path):
        return int(treeObject["mode"],8)

  def _getSize(self,path):
    """
    >>> from subuserlib.classes.gitRepository import GitRepository
    >>> gitRepository = GitRepository(subuserlib.classes.gitRepository.getUser(),subuserlib.classes.gitRepository.hashtestDir)
    >>> fileStructure = gitRepository.getFileStructureAtCommit("master")
    >>> print(fileStructure.getSize("./blah"))
    9
    """
    allObjects = self.lsTree()
    for treeObject in allObjects:
      if os.path.normpath(treeObject["path"]) == os.path.normpath(path):
        return int(treeObject["size"],10)

  def isLegalSymlink(self,path):
    """
    There shouldn't be any risk involved with symlinks in git,
    since git already reads only files which have been checked in.
    """
    return True

class GitException(Exception):
  pass



# -*- coding: utf-8 -*-

#external imports
from collections import OrderedDict
#internal imports
import subuserlib.verify
import subuserlib.print
import subuserlib.classes.docker.dockerDaemon as dockerDaemon

def getInstalledImagesThatAreInUse(user):
  """
  Returns a dictionary of currently installed images that are currently used by a subsuser directly, or indirectly (as a dependency of another image).

  Returns {imageId(string) : InstalledImage}
  """
  installedImagesThatAreInUse = OrderedDict() # {imageId : installedImage}
  for _,subuser in user.registry.subusers.items():
    if subuser.imageId:
      try:
        installedImage = subuser.user.installedImages[subuser.imageId]
        for inUseInstalledImage in installedImage.getImageLineage():
          installedImagesThatAreInUse[inUseInstalledImage.imageId] = inUseInstalledImage
      except KeyError:
        user.registry.log("Warning: No image for %s installed."%subuser.name)
  return installedImagesThatAreInUse

def removeOldImages(user,dryrun=False,yes=False,sourceRepo=None,imageSourceName=None):
  installedImagesThatAreInUse = getInstalledImagesThatAreInUse(user)
  imagesToBeRemoved = []
  for installedImageId,installedImage in user.installedImages.items():
    if (   (not installedImageId in installedImagesThatAreInUse)
       and (not sourceRepo or installedImage.sourceRepoId == sourceRepo.id)
       and (not imageSourceName or installedImage.imageSourceName == imageSourceName)):
      imagesToBeRemoved.append(installedImage)
  if not imagesToBeRemoved:
    user.registry.log("There are no unused images to be removed.")
    return
  user.registry.log("The following images are uneeded and would be deleted.")
  user.registry.log("DOCKER-ID : SUBUSER-ID")
  # List images to be removed
  for installedImage in imagesToBeRemoved:
    user.registry.log("Removing unneeded image "+installedImage.imageId + " : " + installedImage.imageSource.getIdentifier())
  if dryrun:
    return
  # Ask user if we should continue?
  try:
    user.registry.log("Would you like to remove these images now? [Y/n]:",verbosityLevel=4)
    answer = input("Would you like to remove these images now? [Y/n]:")
    removeImages = not (answer == "n")
  except EOFError: # When not running interactively...
    user.registry.log("")
    removeImages = True
  if yes or removeImages:
    # This is a very basic emulation of a topological sort,
    # so we can remove images which depend on eachother.
    nextRound = imagesToBeRemoved
    didSomething = True
    while nextRound and didSomething:
      thisRound = nextRound
      nextRound = []
      didSomething = False
      for installedImage in thisRound:
        installedImage.removeCachedRuntimes()
        try:
          installedImage.removeDockerImage()
          didSomething = True
        except dockerDaemon.ContainerDependsOnImageException:
          nextRound.append(installedImage)
    subuserlib.verify.verify(user)
    user.registry.commit()



# -*- coding: utf-8 -*-

"""
Each subuser has a set of permissions which specify what parts of the host system it is allowed to access.
"""

#external imports
import collections
import hashlib
import sys
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
from subuserlib.classes.fileBackedObject import FileBackedObject
import subuserlib.permissions

class Permissions(collections.OrderedDict,UserOwnedObject,FileBackedObject):
  def __init__(self,user,initialPermissions,writePath=None):
    self.writePath = writePath
    UserOwnedObject.__init__(self,user)
    collections.OrderedDict.__init__(self)
    self.update(initialPermissions)

  def getHash(self):
    """
    Return the SHA512 hash of the given permissions.
    """
    hasher = hashlib.sha512()
    hasher.update(self.json.encode('utf-8'))
    return hasher.hexdigest()

  def applyChanges(self,permissionsToRemove,permissionsToAddOrChange):
    defaults = subuserlib.permissions.getDefaults()
    for permission in permissionsToRemove:
      self[permission] = defaults[permission]
    for permission,value in permissionsToAddOrChange.items():
      self[permission] = value

  def save(self,_have_lock=False):
    if (not self.user._has_lock) and (not _have_lock):
      raise Exception("Programmer error. Saving permissions without first aquiring lock! Please report this incident to: https://github.com/subuser-security/subuser/issues")
    with self.user.endUser.get_file(self.writePath,'w') as fd:
      fd.write(subuserlib.permissions.getJSONString(self))

  @property
  def description(self):
    return subuserlib.permissions.getDescription(self)

  def describe(self,rst=False):
    self.user.registry.log(subuserlib.permissions.getDescription(self,rst=rst))

  @property
  def json(self):
    return subuserlib.permissions.getJSONString(self)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#external imports
import sys
import optparse
import os
#internal imports
from subuserlib.classes.user import User
import subuserlib.commandLineArguments
import subuserlib.profile

##############################################################
def parseCliArgs(sysargs):
  usage = "usage: subuser run [arguments-for-run-command] SUBUSER-NAME [arguments-for-subuser]"
  description = """Run the given subuser.

For example:

    $ subuser run iceweasel

Will launch the subuser named iceweasel

To launch the subuser with another executable than it's default use the ``--entrypoint`` option.

If the SUBUSER_EXTRA_DOCKER_ARGS environment variable is set. Those arguments will be passed to Docker.
"""
  parser=optparse.OptionParser(usage=usage,description=description,formatter=subuserlib.commandLineArguments.HelpFormatterThatDoesntReformatDescription())
  parser.add_option("--entrypoint", dest="entrypoint",default=None,help="Override the default executable for this subuser.")
  parser.add_option("--dry", dest="dry",action="store_true",default=False,help="Dry run, only display what commands would be called by subuser, don't actually run anything.")
  parser.add_option("--dry-one-arg-per-line", dest="dry_one_arg_per_line",action="store_true",default=False,help="Dry run, only display what commands would be called by subuser, don't actually run anything. Displays one argument per line when displaying the commands that would be calleed by subuser, in order to improve the ability for the test suit to detect changes to the command using a line based diff.")
  return parser.parse_args(args=sysargs)

class ArgParser():
  def __init__(self):
    self.preArgs = []
    self.subuserName = None
    self.subuserArgs = []
    self.consumedSubuserName = False

  def readArg(self,arg):
    if not self.consumedSubuserName:
      if arg.startswith("-"):
        self.preArgs.append(arg)
      else:
        self.subuserName = arg
        self.consumedSubuserName = True
    else:
      self.subuserArgs.append(arg)

#################################################################################################

@subuserlib.profile.do_cprofile
def runCommand(args):
  preArgs = []
  argParser = ArgParser()
  for arg in args:
    argParser.readArg(arg)

  if not argParser.consumedSubuserName:
    print("Subuser name not listed.")
    parseCliArgs(["--help"])

  options,_ = parseCliArgs(argParser.preArgs)

  user = User()
  if not "SUBUSER_VERBOSITY" in os.environ:
    user.registry.logOutputVerbosity = 0
  if argParser.subuserName in user.registry.subusers:
    try:
      extraDockerFlags = os.environ["SUBUSER_EXTRA_DOCKER_ARGS"].split(" ")
    except KeyError:
      extraDockerFlags = []
    try:
      subuser = user.registry.subusers[argParser.subuserName]
      runtime = subuser.getRuntime(os.environ,extraDockerFlags=extraDockerFlags,entrypoint=options.entrypoint)
      if runtime:
        if options.dry or options.dry_one_arg_per_line:
          if subuser.imageId:
            print("The image will be prepared using the Dockerfile:")
            print(subuser.getRunReadyImage().generateImagePreparationDockerfile())
            print("The command to launch the image is:")
            if options.dry:
              print(runtime.getPrettyCommand(argParser.subuserArgs))
            elif options.dry_one_arg_per_line:
              print("docker \\")
              print("  `"+"` \\\n  `".join(runtime.getCommand(argParser.subuserArgs))+"`")
          else:
            sys.exit("There is no installed image for this subuser. Cannot run.")
        else:
          runtime.run(argParser.subuserArgs)
      else:
        sys.exit("The subuser's image failed to build. Please use the subuser registry log and subuser repair commands for more information.")
    except (subuserlib.classes.subuser.SubuserHasNoPermissionsException,subuserlib.classes.subuserSubmodules.run.runtimeCache.NoRuntimeCacheForSubusersWhichDontHaveExistantImagesException) as e:
      sys.exit(str(e))
  else:
    sys.exit("Subuser "+argParser.subuserName + " not found.\nUse --help for help.")



# -*- coding: utf-8 -*-

"""
Each user's settings are stored in a "registry". This is a git repository with a set of json files which store the state of the subuser installation.
"""

#external imports
import os
import errno
import sys
from contextlib import contextmanager
import json
#internal imports
from subuserlib.classes import repositories
from subuserlib.classes import subusers
from subuserlib.classes import userOwnedObject
from subuserlib.classes.gitRepository import GitRepository
import subuserlib.print
import subuserlib.executablePath

class Registry(userOwnedObject.UserOwnedObject):
  def __init__(self,user,gitReadHash="master", ignoreVersionLocks=False, initialized = False):
    self.__subusers = None
    self.__changeLog = u""
    self.commit_message = None
    self.__changed = False
    self.logOutputVerbosity = 2
    self.initialized = initialized
    self.lastVerbosityLevel = None
    self.ignoreVersionLocks = ignoreVersionLocks
    if "SUBUSER_VERBOSITY" in os.environ:
      try:
        self.logOutputVerbosity = int(os.environ["SUBUSER_VERBOSITY"])
      except ValueError:
        subuserlib.print.printWithoutCrashing("Invalid verbosity setting! Verbosity may be set to any integer.")
    self.__repositories = None
    self.gitRepository = None
    self.gitReadHash = gitReadHash
    userOwnedObject.UserOwnedObject.__init__(self,user)
    self.registryDir = self.user.config["registry-dir"]
    self.logFilePath = os.path.join(self.registryDir,"commit_log")
    self.gitRepository = GitRepository(self.user,self.registryDir)

  @property
  def subusers(self):
    if self.__subusers is None:
      self.__subusers = subusers.Subusers(self.user)
    return self.__subusers

  @property
  def repositories(self):
    if not self.__repositories:
      self.__repositories = repositories.Repositories(self.user)
    return self.__repositories

  def ensureGitRepoInitialized(self):
    if not os.path.exists(os.path.join(self.user.config["registry-dir"],".git")):
      self.initialized = False
      # Ensure git is setup before we start to make changes.
      self.gitRepository.assertGitSetup()
      self.user.endUser.makedirs(self.user.config["registry-dir"])
      self.gitRepository.run(["init"])
      self.logChange("Initial commit.")
      self.commit("Initial commit.",_no_lock_needed = True)
    self.initialized = True

  def log(self,message,verbosityLevel=1,notify=False):
    """
    If the current verbosity level is equal to or greater than verbosityLevel, print the message to the screen.
    If the current verbosity level is equal to or greater than verbosityLevel minus one, add the message to the log.
    Do not mark the registry as changed.
    The notify option will create a popup dialog with the message if the notify-send command exists.
    """
    message = message.rstrip()
    if (verbosityLevel-1) <= self.logOutputVerbosity:
      self.__changeLog = self.__changeLog + message + u"\n"
    if self.lastVerbosityLevel == 2 and self.logOutputVerbosity == 2 and sys.stdout.isatty():
      CURSOR_UP_ONE = '\x1b[1A'
      ERASE_LINE = '\x1b[2K'
      print(CURSOR_UP_ONE + ERASE_LINE + CURSOR_UP_ONE)
    if verbosityLevel <= self.logOutputVerbosity:
      subuserlib.print.printWithoutCrashing(message)
      if notify and subuserlib.executablePath.which("notify-send"):
        self.user.endUser.call(["notify-send",message])
    self.lastVerbosityLevel = verbosityLevel

  def logChange(self,message,verbosityLevel=1):
    """
    Add a log message to the registry's change log, and mark the registry as changed.
    """
    self.log(message,verbosityLevel=verbosityLevel)
    self.__changed = True

  def setChanged(self,changed=True):
    self.__changed = changed

  def logRenameCommit(self, message):
    """
    Add a new message to the top of the log.
    """
    self.__changeLog = message + u"\n" + self.__changeLog

  def commit(self,message=None,_no_lock_needed=False):
    """
    Git commit the changes to the registry files, installed-miages.json and subusers.json.
    """
    if (not self.user._has_lock) and (not _no_lock_needed):
      sys.exit("Programmer error. Committing to registry without first aquiring lock! Please report this incident to: https://github.com/subuser-security/subuser/issues")
    if self.__changed:
      self.repositories.save()
      self.subusers.save()
      with self.user.endUser.get_file(self.logFilePath) as fd:
        fd.write(self.__changeLog)
      self.gitRepository.run(["add","."])
      if message is None:
        if self.commit_message is not None:
          message = self.commit_message
        else:
          message = self.__changeLog
      self.gitRepository.commit(message)
      # Log to live log
      announcement = {}
      announcement["commit"] = self.gitRepository.getHashOfRef("master")
      self.logToLiveLog(announcement)
      self.__changed = False
      self.__changeLog = u""

  def logToLiveLog(self,announcement):
    announcementJson = json.dumps(announcement)
    liveLogDir=os.path.join(self.user.endUser.homeDir,".subuser/registry-live-log")
    if os.path.isdir(liveLogDir):
      for liveLogPid in os.listdir(liveLogDir):
        liveLogPath = os.path.join(liveLogDir,liveLogPid)
        try:
          liveLog = os.open(liveLogPath,os.O_WRONLY|os.O_NONBLOCK)
          os.write(liveLog,announcementJson)
        except OSError:
          pass
        # TODO Note: We don't close the file descriptors, because doing so makes the pipe close on the other end too. This would be a file descriptor leak if this method was used in any long running process(which it is not).

  def cleanOutOldPermissions(self):
    for permissions_folder_name in self.gitRepository.getFileStructureAtCommit(self.gitReadHash).lsFolders("permissions"):
      exists = os.path.exists(os.path.join(self.registryDir,"permissions",permissions_folder_name))
      if exists and permissions_folder_name not in self.subusers:
        self.logChange("Removing left over permissions for no-longer extant subuser %s"%permissions_folder_name,2)
        try:
          self.gitRepository.run(["rm","-r",os.path.join("permissions",permissions_folder_name)])
        except subuserlib.classes.gitRepository.GitException as e:
          self.log(" %s"%str(e))



# -*- coding: utf-8 -*-

"""
A repository is a collection of ``ImageSource`` s which are published in a git repo.
"""

#external imports
import os
import shutil
import json
from collections import OrderedDict
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
from subuserlib.classes.imageSource import ImageSource
from subuserlib.classes.describable import Describable
from subuserlib.classes.gitRepository import GitRepository
from subuserlib.classes.fileStructure import BasicFileStructure
import subuserlib.version

class Repository(OrderedDict,UserOwnedObject,Describable):
  def __init__(self,user,name,gitOriginURI=None,gitCommitHash=None,temporary=False,sourceDir=None):
    """
    Repositories can either be managed by git, or simply be normal directories on the user's computer. If ``sourceDir`` is not set to None, then ``gitOriginURI`` is ignored and the repository is assumed to be a simple directory.
    """
    self.name = name
    self.gitOriginURI = gitOriginURI
    self.gitCommitHash = gitCommitHash
    self.temporary = temporary
    self.sourceDir = sourceDir
    self.__fileStructure = None
    UserOwnedObject.__init__(self,user)
    super().__init__()
    self.gitRepository = GitRepository(user,self.repoPath)
    if not self.isPresent():
      self.updateSources(initialUpdate=True)
    if self.isPresent():
      self.repoConfig = self.loadRepoConfig()
      self.loadImageSources()

  @property
  def uri(self):
    if self.isLocal:
      return self.sourceDir
    else:
      return self.gitOriginURI

  @property
  def fileStructure(self):
    if self.__fileStructure is None:
      if self.isLocal:
        self.__fileStructure = BasicFileStructure(self.sourceDir)
      else:
        if self.gitCommitHash is None:
          self.updateGitCommitHash()
        self.__fileStructure = self.gitRepository.getFileStructureAtCommit(self.gitCommitHash)
    return self.__fileStructure

  @property
  def displayName(self):
    """
    How should we refer to this repository when communicating with the user?
    """
    if self.temporary:
      if self.isLocal:
        return self.sourceDir
      else:
        return self.gitOriginURI
    else:
      return self.name

  def describe(self):
    print("Repository: "+self.displayName)
    print("------------")
    if self.isLocal:
      print("Is a local(non-git) repository.")
      if not self.temporary:
        print("Located at: " + self.repoPath)
    if self.temporary:
      print("Is a temporary repository.")
    else:
      if not self.isLocal:
        print("Cloned from: "+self.gitOriginURI)
        print("Currently at commit: "+self.gitCommitHash)

  def getSortedList(self):
    """
    Return a list of image sources sorted by name.
    """
    return list(sorted(self.values(),key=lambda imageSource:imageSource.name))

  @property
  def repoPath(self):
    """ Get the path of the repo's sources on disk. """
    if self.isLocal:
      return self.sourceDir
    else:
      return os.path.join(self.user.config["repositories-dir"],self.name)

  def loadRepoConfig(self):
    """
    Either returns the config as a dictionary or None if no configuration exists or can be parsed.
    """
    def verifyPaths(dictionary,paths):
      """
      Looks through a dictionary at all entries that can be considered to be paths, and ensures that they do not contain any relative upwards references.
      Throws a ValueError if they do.
      """
      for path in paths:
        if path in dictionary and path.startswith("../") or "/../" in path:
          raise ValueError("Paths in .subuser.json may not be relative to a higher directory.")
    if self.fileStructure.exists("./.subuser.json"):
      configFileContents = self.fileStructure.read("./.subuser.json")
    else:
      return None
    repoConfig = json.loads(configFileContents)
    # Validate untrusted input
    verifyPaths(repoConfig,["image-sources-dir"])
    if "explicit-image-sources" in repoConfig:
      for explicitImageSource in repoConfig["explicit-image-sources"]:
        verifyPaths(explicitImageSource,["image-file","permissions-file","build-context"])
    return repoConfig

  @property
  def imageSourcesDir(self):
    """
    Get the path of the repo's subuser root on disk on the host.
    """
    return os.path.join(self.repoPath,self.relativeImageSourcesDir)

  @property
  def relativeImageSourcesDir(self):
    """
    Get the path of the repo's subuser root on disk on the host.
    """
    repoConfig = self.repoConfig
    if repoConfig and "image-sources-dir" in repoConfig:
      return repoConfig["image-sources-dir"]
    else:
      return "./"

  def isInUse(self):
    """
    Are there any installed images or subusers from this repository?
    """
    for _,installedImage in self.user.installedImages.items():
      if self.name == installedImage.sourceRepoId:
          return True
      for _,subuser in self.user.registry.subusers.items():
        try:
          if self.name == subuser.imageSource.repo.name:
            return True
        except subuserlib.classes.subuser.NoImageSourceException:
          pass
    return False

  @property
  def isLocal(self):
    if self.sourceDir:
      return True
    return False

  def removeGitRepo(self):
    """
    Remove the downloaded git repo associated with this repository from disk.
    """
    if not self.isLocal:
      shutil.rmtree(self.repoPath)

  def isPresent(self):
    """
    Returns True if the repository's files are present on the system. (Cloned or local)
    """
    return os.path.exists(self.repoPath)

  def updateSources(self,initialUpdate=False):
    """
    Pull(or clone) the repo's ImageSources from git origin.
    """
    if self.isLocal:
      return
    if not self.isPresent():
      new = True
      self.user.registry.log("Cloning repository "+self.name+" from "+self.gitOriginURI)
      if self.gitRepository.clone(self.gitOriginURI) != 0:
        self.user.registry.log("Clone failed.")
        return
    else:
      new = False
    try:
      self.gitRepository.run(["fetch","--all"])
    except Exception: # For some reason, git outputs normal messages to stderr.
      pass
    if self.updateGitCommitHash():
      if not new:
        self.user.registry.logChange("Updated repository "+self.displayName)
      if not initialUpdate:
        self.loadImageSources()

  def serializeToDict(self):
    """
    Return a dictionary which describes the image sources available in this repository.
    """
    imageSourcesDict = {}
    for name,imageSource in self.items():
      imageSourcesDict[name] = imageSource.serializeToDict()
    return imageSourcesDict

  def loadImageSources(self):
    """
    Load ImageSources from disk into memory.
    """
    imageNames = self.fileStructure.lsFolders(self.relativeImageSourcesDir)
    for imageName in imageNames:
      imageSource = ImageSource(self.user,self,imageName)
      if self.fileStructure.exists(imageSource.getRelativePermissionsFilePath()):
        self[imageName] = imageSource
    if self.repoConfig is not None and "explicit-image-sources" in self.repoConfig:
      for imageName,config in self.repoConfig["explicit-image-sources"].items():
        assert config is not None
        self[imageName] = ImageSource(self.user,self,imageName,explicitConfig = config)

  def updateGitCommitHash(self):
    """
    Update the internally stored git commit hash to the current git HEAD of the repository.
    Returns True if the repository has been updated.
    Otherwise false.
    """
    if self.isLocal:
      return True
    # Default
    master = "refs/remotes/origin/master"
    newCommitHash = self.gitRepository.getHashOfRef(master)
    # First we check for version constraints on the repository.
    if self.gitRepository.getFileStructureAtCommit(master).exists("./.subuser.json"):
      configFileContents = self.gitRepository.getFileStructureAtCommit(master).read("./.subuser.json")
      configAtMaster = json.loads(configFileContents)
      if "subuser-version-constraints" in configAtMaster:
        versionConstraints = configAtMaster["subuser-version-constraints"]
        subuserVersion = subuserlib.version.getSubuserVersion(self.user)
        for constraint in versionConstraints:
          if not len(constraint) == 3:
            raise SyntaxError("Error in .subuser.json file. Invalid subuser-version-constraints."+ str(versionConstraints))
          op,version,commit = constraint
          from operator import lt,le,eq,ge,gt
          operators = {"<":lt,"<=":le,"==":eq,">=":ge,">":gt}
          try:
            matched = operators[op](subuserVersion,version)
          except KeyError:
            raise SyntaxError("Error in .subuser.json file. Invalid subuser-version-constraints.  \""+op+"\" is not a valid operator.\n\n"+ str(versionConstraints))
          if matched:
            try:
              newCommitHash = self.gitRepository.getHashOfRef("refs/remotes/origin/"+commit)
            except OSError as e:
              if len(commit) == 40:
                newCommitHash = commit
              else:
                raise e
            break
        else:
          raise SyntaxError("Error reading .subuser.json file, no version constraints matched the current subuser version ("+subuserVersion+").\n\n"+str(versionConstraints))
    updated = not (newCommitHash == self.gitCommitHash)
    self.gitCommitHash = newCommitHash
    self.__fileStructure = None
    return updated



# -*- coding: utf-8 -*-

"""
The ``User`` object is the base object which owns all other objects in a running subuser instance.
"""

#external imports
import getpass
import os
import sys
import pwd
import errno
#internal imports
from subuserlib.classes import registry
from subuserlib.classes import config
from subuserlib.classes import installedImages
from subuserlib.classes.docker import dockerDaemon
from subuserlib.classes.endUser import EndUser
from subuserlib import test
from subuserlib import paths
import subuserlib.lock

class User(object):
  """
  This class provides a "base" User object used by subuser.  This is the stem of a tree like data structure which holds all of the various objects owned by a given user.

  You create a new User object by passing the username and home dir of the user.

  >>> import subuserlib.classes.user
  >>> u = subuserlib.classes.user.User(name="root",homeDir="/root/")
  >>> u.homeDir
  '/root/'
  """
  def __init__(self,name=None,homeDir=None,_locked=False):
    self.__config = None
    self.__registry = None
    self.__installedImages = None
    self.__dockerDaemon = None
    self.__runtimeCache = None
    self._has_lock = _locked
    self.name = name
    if homeDir:
      self.homeDir = homeDir
    elif test.testing:
      self.homeDir = os.getcwd()
    else:
      self.homeDir = os.path.expanduser("~")
    self.endUser = EndUser(self)

  @property
  def config(self):
    """
    Get the user's :doc:`Config <config>` object.

    Note: the user's config will be loaded the first time this is called.
    """
    if self.__config == None:
      self.__config = config.Config(self)
    return self.__config

  @property
  def registry(self):
    """
    Get the user's subuser :doc:`Registry <registry>`.

    Note: the registry will be loaded the first time this is called.
    """
    if self.__registry == None:
      self.__registry = registry.Registry(self)
      self.__registry.ensureGitRepoInitialized()
    return self.__registry

  @registry.setter
  def registry(self, registry):
    self.__registry = registry

  def reloadRegistry(self):
    """
    Reload registry from disk.
    """
    self.__registry = None

  @property
  def installedImages(self):
    """
    Get the user's  :doc:`InstalledImages <installed-images>` list.

    Note: the installed images list will be loaded the first time this is called.
    """
    if self.__installedImages == None:
      self.__installedImages = installedImages.InstalledImages(self)
    return self.__installedImages

  @property
  def dockerDaemon(self):
    """
    Get the :doc:`DockerDaemon <docker>` object.  You will use this to communicate with the Docker daemon.
    """
    if self.__dockerDaemon == None:
      self.__dockerDaemon = dockerDaemon.DockerDaemon(self)
    return self.__dockerDaemon

class LockedUser():
  def __init__(self,name=None,homeDir=None):
    self.lock = None
    self.__user = User(name=name,homeDir=homeDir,_locked=True)

  def __enter__(self):
    try:
      self.__user.endUser.makedirs(self.__user.config["lock-dir"])
    except OSError as exception:
      if exception.errno != errno.EEXIST:
        raise
    try:
      self.lock = subuserlib.lock.getLock(self.__user.endUser.get_file(os.path.join(self.__user.config["lock-dir"],"registry.lock"),'w'),timeout=1)
      self.lock.__enter__()
    except IOError as e:
      if e.errno != errno.EINTR:
        raise e
      sys.exit("Another subuser process is currently running and has a lock on the registry. Please try again later.")
    return self.__user

  def __exit__(self, type, value, traceback):
    self.lock.__exit__()



#!/usr/bin/python3
import sys,subprocess
sys.path.append("../../logic")
import subuserlib.permissions

with open("permissions-dot-json-file-format.rst","w") as permissions_format_file:
  permissions_format_file.write(subuserlib.permissions.getDocs())



# -*- coding: utf-8 -*-

"""
Stores metadata about images which are built to encorporate changes to subuser images which are required in order to implement various permissions.
"""

#external imports
import os
import json
from collections import OrderedDict
#internal imports
from subuserlib.classes.userOwnedObject import UserOwnedObject
from subuserlib.classes.fileBackedObject import FileBackedObject

class RuntimeCache(dict,UserOwnedObject,FileBackedObject):
  def __init__(self,user,subuser):
    self.subuser = subuser
    UserOwnedObject.__init__(self,user)
    self.load()

  @property
  def pathToCurrentImagesRuntimeCacheDir(self):
    return os.path.join(self.user.config["runtime-cache"],self.subuser.imageId)

  @property
  def runtimeCacheFilePath(self):
    return os.path.join(self.pathToCurrentImagesRuntimeCacheDir,self.subuser.getRunReadyImage().sourceHash+".json")

  def save(self):
    try:
      self.user.endUser.makedirs(self.pathToCurrentImagesRuntimeCacheDir)
    except OSError:
      pass
    with self.user.endUser.get_file(self.runtimeCacheFilePath,mode='w') as runtimeCacheFileHandle:
      json.dump(self,runtimeCacheFileHandle,indent=1,separators=(',',': '))

  def reload(self):
    self.save()
    self.load()

  def load(self):
    if not self.subuser.imageId:
      raise NoRuntimeCacheForSubusersWhichDontHaveExistantImagesException("No runnable image for subuser "+self.subuser.name+" found. Use\n\n $ subuser repair\n\nTo repair your instalation.")
    runtimeCacheFilePath = self.runtimeCacheFilePath
    if os.path.exists(runtimeCacheFilePath):
      with open(runtimeCacheFilePath,mode="r") as runtimeCacheFileHandle:
        runtimeCacheInfo = json.load(runtimeCacheFileHandle, object_pairs_hook=OrderedDict)
        self.update(runtimeCacheInfo)

class NoRuntimeCacheForSubusersWhichDontHaveExistantImagesException(Exception):
  pass



# -*- coding: utf-8 -*-

"""
Module used for the loading and saving of permissions.json files. Contains the default permissions list.
"""

#external imports
import json
from collections import OrderedDict
import sys
import os
import copy
#internal imports
# import ...
allImagesMustHavePermissions = "All subuser images must have a permissions.json file as defined by the permissions.json standard: <http://subuser.org/subuser-standard/permissions-dot-json-file-format.html>"

class ListOfStrings():
  pass

class StringToStringDict():
  pass

class StringToCommandDict():
  pass

def validateUserDirs(userDirs):
  for userDir in userDirs:
    if ".." in userDir or userDir.startswith("./"):
      raise SyntaxError("Error, user dir permissions may not contain relative paths. The user dir: \""+userDir+"\" is forbidden.")
    if userDir.startswith("/"):
      raise SyntaxError("Error, user dir permissions may not contain system wide absolute paths. The user dir: \""+userDir+"\" is forbidden.")

# Defaults from http://subuser.org/subuser-standard/permissions-dot-json-file-format.html
# This is a comprehensive list of all permissions
supportedPermissions = OrderedDict([
#The format of this dictionary is
#(NAME_OF_DANGER_LEVEL,DESCRIPTION_OF_DANGER_LEVEL):
#  [("<name-of-permission>",
#    [("describe",<lambda that generates english langauge description of permission, given its value.>)
#    ,("default",<default value of permission>)
#    ,("types",[<valid types that the permissions value can have, used for permission file validation>])
#    ,("description","""Description meant to show up in standard manaual."""
#    OPTIONALLY
#    ,("subpermissions",
#      [
#      <same format as for top level permissions>
#      ]
#    ,("conflicts",[<permissions that cannot be granted at the same time as this permission>])
#    ]
#  ]

(("prelude","Prelude:"),

  OrderedDict([("description",
    {"describe":lambda v: v
    ,"description":"""This field describes what the subuser is/what it does.

  Ex::

    "description"                : "Simple universal text editor."
"""
    ,"default":""
    ,"types":[str]})

  ,("maintainer",
    {"describe":lambda maintainer : maintainer
    ,"description":"""This field marks who is responsible for the ``permissions.json`` file, and accompanying ``Dockerfile``.  It does NOT mark who is responsible for the image itself.

  Ex::

    ,"maintainer"                : "Timothy Hobbs <timothyhobbs (at) seznam dot cz>"
"""
    ,"default":""
    ,"types":[str]})

  ,("executable",
    {"describe":lambda executable : executable if executable else "Is a library."
    ,"description":"""This field denotes the absolute path within the Docker image where the given image's executable resides. This value is optional. if it is not present, than the subuser image cannot be run (but may be depended upon by other subuser images).

  Ex::

    ,"executable"                : "/usr/bin/vim"
"""
    ,"default":None
    ,"types":[type(None),str,ListOfStrings]})

  ,("entrypoints",
    {"describe":lambda entrypoints: "'"+ "', '".join(entrypoints.keys())+"'" if entrypoints else ""
    ,"description":"""This optional feild allows you to add "entrypoints" to your subuser. These are executables that can be added, if the user so wishes, to the PATH on the host system. This is a dictionary which maps "desired name on host" to "path within subuser image".

  Ex::

    ,"entrypoints"                : {"mk":"/usr/bin/mk","cc","/usr/local/bin/cc"}
"""
    ,"default":None
    ,"types":[type(None),StringToCommandDict]})
  ])

),(("conservative","Conservative permissions(These are safe):"),OrderedDict([

  ("basic-common-permissions",
    {"describe":lambda v: "To access basic information such as timezone and local and to save data to its own homedir." if v is True else ""
    ,"description":"""Grant access to asic information such as timezone and local and to save data to its own homedir.

  Ex::

    ,"basic-common-permissions"                : true


  Ex1::


    ,"basic-common-permissions"                : {"stateful-home":true,"inherit-timezone":false}

"""
    ,"default":False
    ,"types":[bool,OrderedDict]
    ,"subpermissions-batch-set":True
    ,"subpermissions":OrderedDict([

      ("stateful-home",
        {"describe":lambda p : "To have its own home directory where it can save files and settings." if p else ""
        ,"description":"""Changes that the subuser makes to it's home directory should be saved to a special subuser-homes directory.
"""
        ,"default":False
        ,"types":[bool]})

      ,("inherit-locale",
        {"describe":lambda p : "To find out which language you speak and what region you live in." if p else ""
        ,"description":"""Automatically set the $LANG and $LANGUAGE environment variable in the container to the value outside of the container. Note: You do not have to set this if you have set ``basic-common-permissions``.
"""
        ,"default":False
        ,"types":[bool]})

      ,("inherit-timezone",
        {"describe": lambda p : "To find out your current timezone." if p else ""
        ,"description":"""Automatically set the $TZ environment variable in the container to the value outside of the container.  Give the sub user read only access to the ``/etc/localtime`` file. Note: You do not have to set this if you have set ``basic-common-permissions``.
"""
        ,"default":False
        ,"types":[bool]})
      ])
    })

  ,("memory-limit",
    {"describe": lambda p: "Memory limited to %s"%p if p else ""
    ,"description":"""Limit subuser's available memory to this value. This is a string of format "100m" or "2g", where "m" is megabytes and "g" is gigabytes.

  Ex::

     "memory-limit":"200m"
"""
    ,"default":None
    ,"types":[type(None),str]})

  ,("max-cpus",
    {"describe":lambda p: "CPUs limited to %s"%p if p else ""
    ,"description":"""Limit number of CPUs subuser can use. This is a float.

  Ex::

    "max-cpus":2.0
"""
    ,"default":None
    ,"types":[type(None),float]})
  ])

),(("moderate","Moderate permissions(These are probably safe):"),OrderedDict([

  ("gui",
    {"describe": lambda guiOptions : "To be able to display windows." if guiOptions else ""
    ,"description":"""Is the subuser allowed to display a graphical user interface?

  Ex (enable gui, but do not enable extra fetures such as clipboard or system-tray icons)::

     "gui":true

  Ex1 (enable gui along with select features)::

    "gui":{"clipboard":true}
"""
    ,"default":False
    ,"types":[bool,OrderedDict]
    ,"conflicts":["x11"]
    ,"subpermissions": OrderedDict([

      ("clipboard",
        {"describe":lambda p : "To be able to access the host's clipboard." if p else ""
        ,"description":"""Is the subuser allowed to read and write to the clipboard?
"""
        ,"default":False
        ,"types":[bool]})

      ,("system-tray",
        {"describe":lambda p : "To be able to create system tray icons." if p else ""
        ,"description":"""Is this subuser allowed to display system-tray icons?
"""
        ,"default":False
        ,"types":[bool]})

      ,("cursors",
        {"describe":lambda p : "To be able to change the mouse's cursor icon." if p else ""
        ,"description":"""Is this subuser allowed to change the way the mouse cursor is displayed?
"""
        ,"default":False
        ,"types":[bool]})

      ,("border-color",
        {"describe":lambda p : "The border color will be "+p if p else "Warning: There will be no border coloring."
        ,"description":"""Set the border color for the window. This allows you to easily distinguish between untrusted windows and trusted ones.

  Ex (set border color to green)::

    "gui":{"border-color":"green"}

  Ex1 (display windows without adding special border color.)::

    "gui":{"border-color":nil}

"""
        ,"default":"red"
        ,"types":[type(None),str]})
      ])
    })

  ,("user-dirs",
    {"describe":lambda userDirs : "To access to the following user directories: '~/"+"' '~/".join(userDirs)+"'" if userDirs else ""
    ,"description":"""A list of relative paths to user directories which are to be shared between the host and the given image. The subuser is given read-write access to any user directories listed.

  Ex::

     ,"user-dirs"                 : ["Downloads"]

  In this example, the subuser is able to access the ``~/Downloads`` directory on the host by visiting the ``~/Userdirs/Downloads`` directory within the container.
"""
    ,"default":[]
    ,"validate": validateUserDirs
    ,"types":[ListOfStrings]})

  ,("inherit-envvars",
    {"describe":lambda envVars : "To access to the following environment variables: "+" ".join(envVars) if envVars else ""
    ,"description":"""A list of environment variables which the image will inherit from the host environment when started.

  Ex::

     ,"inherit-envvars"           : ["PGUSER","PGHOST"]
"""
    ,"default":[]
    ,"types":[ListOfStrings]})

  ,("pulseaudio",
    {"describe":lambda p : "To access to your pulseaudio server, can play sounds/record sound." if p else ""
    ,"description":"""The subuser is allowed to access the pulseaudio server on the host.

  .. warning:: This means, not only can the subuser play sounds, but it may listen to your microphone too!
"""
    ,"default":False
    ,"types":[bool]})

  ,("sound-card",
    {"describe":lambda p : "To access to your soundcard, can play sounds/record sound." if p else ""
    ,"description":"""The subuser is allowed to access the soundcard on the host.

  .. warning:: This means, not only can the subuser play sounds, but it may listen to your microphone too!
"""
    ,"default":False
    ,"types":[bool]})

  ,("webcam",
    {"describe":lambda p : "To access your computer's webcam/can see you." if p else ""
    ,"description":"""The subuser is allowed to access the computer's webcam/USB webcams.
"""
    ,"default":False
    ,"types":[bool]})

  ,("access-working-directory",
    {"describe":lambda p: "To access the directory from which it was launched." if p else ""
    ,"description":"""The subuser is given read-write access to the host user's current working directory.
"""
    ,"default":False
    ,"types":[bool]})

  ,("allow-network-access",
    {"describe":lambda p : "To access the network/internet." if p else ""
    ,"description":"""Should the subuser be allowed to access the network/internet?
"""
    ,"default":False
    ,"types":[bool]})

 ])

),(("liberal","Liberal permissions(These may pose a security risk):"),OrderedDict([

  ("x11",
    {"describe":lambda p: "To display X11 windows and interact with your X11 server directly(log keypresses, read over your shoulder, steal your passwords, control your computer ect.)" if p else ""
    ,"description":"""The subuser is allowed to interact with the x11 server on the host.

  .. note:: Known to be insecure!
"""
    ,"default":False
    ,"types":[bool]})

  ,("system-dirs",
    {"describe":lambda systemDirs : " ".join(["To read and write to the host's `"+source+"` directory, mounted in the container as:`"+dest+"`\n" for source,dest in systemDirs.items()])
    ,"description":"""A dictionary of absolute paths to directories which are to be shared between the host and the given image. The subuser is given read-write access to any user directories listed.

  Ex::

     ,"system-dirs"                 : {"/var/log":"/host/var/log"}

  In this example, the subuser is able to access the ``/var/log`` directory on the host by visiting the ``/host/var/log`` directory within the container.
"""
    ,"default":OrderedDict()
    ,"types":[StringToStringDict]})

  ,("graphics-card",
    {"describe":lambda p : "To access your graphics-card directly for OpenGL tasks." if p else "" 
    ,"description":"""The subuser is allowed to access the graphics-card directly(OpenGL).
"""
    ,"default":False
    ,"types":[bool]})

  ,("serial-devices",
    {"describe":lambda p : "To access serial devices such as programmable micro-controllers and modems." if p else ""
    ,"description":"""The subuser is allowed to access serial devices: ``/dev/ttyACM*``, ``/dev/ttyUSB*``, and ``/dev/ttyS*``.
"""
    ,"default":False
    ,"types":[bool]})

  ,("system-dbus",
    {"describe":lambda p: "To talk to the system dbus daemon." if p else ""
    ,"description":"""Should the subuser be allowed to communicate with the system wide dbus daemon?
"""
    ,"default":False
    ,"types":[bool]})

  ,("sudo",
    {"describe": lambda p: "To be allowed to use the sudo command to run subproccesses as root in the container." if p else ""
    ,"description":"""Grant the subuser sudo privileges within the container.
"""
    ,"default":False
    ,"types":[bool]})

  ,("as-root",
    {"describe": lambda p: "To be allowed to run as root within the container." if p else ""
    ,"description":"""Run the subuser as the root user within the container.
"""
    ,"default":False
    ,"types":[bool]})

  ])

),(("anarchistic","Anarchistic permissions\n   WARNING: These permissions give the subuser full access to your system when run."),OrderedDict([
 
  ("privileged",
    {"describe": lambda p: "To have full access to your system.  To even do things as root outside of its container." if p else ""
    ,"description":"""Should the subuser's Docker container be run in ``privileged`` mode?

  .. warning:: Completely insecure!
"""
    ,"default":False
    ,"types":[bool]})

  ,("run-commands-on-host",
    {"describe": lambda p: "To run commands as a normal user on the host system." if p else ""
    ,"description":"""Should the subuser be able to execute commands as the normal user on the host system? If this is enabled, a ``/subuser/execute`` file will be present in the container. Any text appended to this file will be piped to ``/bin/sh`` on the host machine.

  .. warning:: Obviously completely compromises security.
"""
    ,"default":False
    ,"types":[bool]})

  ])
)])

def getDefaults():
  default_permissions = OrderedDict()
  for level,permissions in supportedPermissions.items():
    for permission,value in permissions.items():
      default_permissions[permission] = copy.deepcopy(value["default"])
  return default_permissions

def getTypeDescriptions(types):
  typeDescriptions = {
    str:"string"
   ,type(None):"nil"
   ,bool:"bool"
   ,float:"floating point number"
   ,ListOfStrings:"List of strings"
   ,StringToStringDict:"Dict of strings."
   ,StringToCommandDict:"Dict mapping names to commands."
   ,OrderedDict: "set of subpermissions"}
  typestrings = []
  for t in types:
    if t in typeDescriptions:
      typestrings.append(typeDescriptions[t])
    else:
      typestrings.append(str(t))
  return typestrings

basicCommonPermissions = ["stateful-home","inherit-locale","inherit-timezone"]

def load(permissionsFilePath=None,permissionsString=None, logger = None):
  """
  Return a dictionary of permissions from the given permissions.json file.  Permissions that are not specified are set to their default values.

  Loading permissions doesn't crash horribly and does something at least slightly sensible.

  >>> getNonDefaultPermissions(load(permissionsString="{}"))
  OrderedDict()

  Valid paths are loaded when the user-dirs permission is set.

  >>> print(json.dumps(getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"Downloads\"]}'))))
  {"user-dirs": ["Downloads"]}

  However, funny business is strictly frowned upon.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"..\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: ".." is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"../\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: "../" is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"./../../\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: "./../../" is forbidden.

  Even if it is well hiden behind valid values.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"Downloads\",\"../\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: "../" is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"Downloads\",\"./Foo\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain relative paths. The user dir: "./Foo" is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"user-dirs\":[\"Downloads\",\"/Foo\"]}'))
  Traceback (most recent call last):
    ...
  SyntaxError: Error, user dir permissions may not contain system wide absolute paths. The user dir: "/Foo" is forbidden.

  >>> getNonDefaultPermissions(load(permissionsString='{\"x11\":true,\"gui\":true}'))
  Traceback (most recent call last):
    ...
  SyntaxError: x11 and gui permissions are mutually exclusive.

  """
  if not permissionsString is None:
    try:
      loading=json.loads(permissionsString, object_pairs_hook=OrderedDict)
    except ValueError:
      sys.exit(permissionsString+" is not valid json.  "+allImagesMustHavePermissions)
  else:
    # read permissions.json file
    with open(permissionsFilePath, 'r') as file_f:
      try:
        loading=json.load(file_f, object_pairs_hook=OrderedDict)
      except ValueError:
        raise SyntaxError(permissionsFilePath+" is not valid json.  "+allImagesMustHavePermissions)
  def validate(permission,value,attrs):
    types = attrs["types"]
    if not type(value) in types:
      if type(value) == list and ListOfStrings in types:
        for s in value:
          if not type(s) == str:
            raise SyntaxError("Permission "+permission+" set to invalid value <"+str(value)+">. Expected a list of strings.")
      elif type(value) == OrderedDict and (StringToStringDict in types or StringToCommandDict in types):
        for s in value.values():
          if not type(s) == str:
            if type(s) == list and StringToCommandDict in types:
              continue
            raise SyntaxError("Permission "+permission+" set to invalid value <"+str(value)+">. Expected a string to string dictionary.")
      else:
        typestrings = getTypeDescriptions(types)
        raise SyntaxError("Permisison "+permission+" set to invalid value <"+str(value)+">. Expected "+" or ".join(typestrings))
    if "validate" in attrs:
      attrs["validate"](value)
 
  # Validate that the permissions are supported by this version of subuser.
  loadedPermissions = OrderedDict()
  for level,permissions in supportedPermissions.items():
    for permission,attributes in permissions.items():
      loadedPermissions[permission] = attributes["default"]
      if permission in loading:
        try:
          loading_permission = loading[permission]
        except KeyError:
          continue
        validate(permission,loading_permission,attributes)
        if "conflicts" in attributes:
          for conflict in attributes["conflicts"]:
            if conflict in loading and loading[conflict] and loading_permission:
              raise SyntaxError(conflict+" and "+permission+" permissions are mutually exclusive.")
        if "subpermissions" in attributes and loading_permission:
          loadedPermissions[permission] = OrderedDict()
          for subpermission,subattrs in attributes["subpermissions"].items():
            loadedPermissions[permission][subpermission] = subattrs["default"] #TODO DRY
            if type(loading_permission) == OrderedDict:
              try:
                loading_subpermission = loading_permission[subpermission]
              except KeyError:
                continue
              validate(permission+"."+subpermission,loading_subpermission,subattrs)
              loadedPermissions[permission][subpermission] = loading_subpermission
            elif type(loading_permission) == bool:
              if "subpermissions-batch-set" in attributes and attributes["subpermissions-batch-set"]:
                loadedPermissions[permission][subpermission] = loading_permission
              else:
                loadedPermissions[permission][subpermission] = subattrs["default"]
        else:
          loadedPermissions[permission] = loading_permission
  for permission,value in loading.items():
    if permission not in loadedPermissions and permission not in basicCommonPermissions:
      if logger is not None:
        logger.log("Permission "+permission+" is not supported by this version of subuser.")
  return loadedPermissions

def getNonDefaultPermissions(permissions):
  """
  Returns the dictionary of permissions which are NOT set to their default values.

  >>> import subuserlib.permissions
  >>> permissions = subuserlib.permissions.load(permissionsString='{"x11":true}')
  >>> getNonDefaultPermissions(permissions) == {u'x11': True}
  True
  """
  nonDefaultPermissions = OrderedDict()
  defaults = getDefaults()
  for permission,value in permissions.items():
    if not value == defaults[permission]:
      nonDefaultPermissions[permission] = value
  return nonDefaultPermissions

def getJSONString(permissions):
  """
  Returns the given permissions as a JSON formated string.
  """
  permissionsToSave = getNonDefaultPermissions(permissions)
  return json.dumps(permissionsToSave,indent=1, separators=(',', ': '))

def getDescription(permissionsToDescribe,rst=False):
  descriptionTemplate = {
    "level_heading":"\n%s" if rst else " %s"
   ,"permission_description":"  - %s: %s"
   ,"sub_permission_description":"   * %s"
   }
  description = []
  for (level,levelDescription),permissions in supportedPermissions.items():
    descriptionForThisLevel = []
    describedSomething = False
    descriptionForThisLevel.append(descriptionTemplate["level_heading"]%levelDescription)
    for permission,attributes in permissions.items():
      try:
        permissionValue = permissionsToDescribe[permission]
      except KeyError:
        continue
      if permissionValue:
        describedSomething = True
        descriptionForThisLevel.append(descriptionTemplate["permission_description"]%(permission,attributes["describe"](permissionValue)))
        if "subpermissions" in attributes and type(permissionValue) == OrderedDict:
          if rst:
            descriptionForThisLevel.append("")
          for subpermission,subattributes in attributes["subpermissions"].items():
            if permissionValue[subpermission]:
              descriptionForThisLevel.append(descriptionTemplate["sub_permission_description"]%subattributes["describe"](permissionValue[subpermission]))
          if rst:
            descriptionForThisLevel.append("")
    if describedSomething:
      description += descriptionForThisLevel
  return "\n".join(description)


def compare(oldDefaults,newDefaults,userApproved):
  """
  Analize permission dictionaries for changes.
  First, compare the old defaults to the user approved permissions.
  By doing so, we aquire an understanding of which permissions have been set by the user, and which are still at their default values.
  We will leave the user-set permissions alone.

  Next, we compare the old non-user set permissions to the new defaults.

  The return value is a tuple of the form: ([removed-permisions],{additions/changes})
  """
  return __compare(oldDefaults = getNonDefaultPermissions(oldDefaults), newDefaults = getNonDefaultPermissions(newDefaults), userApproved = getNonDefaultPermissions(userApproved))

def __compare(oldDefaults,newDefaults,userApproved):
  """
  >>> import subuserlib.permissions
  >>> subuserlib.permissions.__compare(oldDefaults={"a":1,"b":2,"c":3,"d":4,"e":5},newDefaults={"a":1,"b":3,"c":4,"f":4},userApproved={"a":1,"b":5,"e":5,"z":7}) == (["e"],{"f":4})
  True

  If the permissions contain an OrderedDict subgroup where the order doesn't match, the comparison still works.

  >>> from collections import OrderedDict
  >>> subuserlib.permissions.__compare(oldDefaults={"a":1,"b":2,"c":3,"d":4,"e":5,"g":OrderedDict([("a",1),("b",2)])},newDefaults={"a":1,"b":3,"c":4,"f":4,"g":OrderedDict([("b",2),("a",1)])},userApproved={"a":1,"b":5,"e":5,"z":7,"g":OrderedDict([("a",1),("b",2)])}) == (["e"],{"f":4})
  True
  >>> subuserlib.permissions.__compare(oldDefaults={"a":1,"b":2,"c":3,"d":4,"e":5,"g":OrderedDict([("a",1),("b",2)])},newDefaults={"a":1,"b":3,"c":4,"f":4,"g":OrderedDict([("b",3),("a",1)])},userApproved={"a":1,"b":5,"e":5,"z":7,"g":OrderedDict([("a",1),("b",2)])}) == (["e"],{"f":4,"g":OrderedDict([("b",3),("a",1)])})
  True
  """
  def neq(a,b):
    if type(a) == type(b) and type(a) == OrderedDict:
      if len(a) != len(b):
        return True
      for key,value in a.items():
        if b[key] != value:
          return True
      return False
    else:
      return a != b
  # This set is the union of the user approved permissions
  userChangedPermissions = set()
  for key,value in userApproved.items():
    try:
      if neq(oldDefaults[key],value):
        userChangedPermissions.add(key)
    except KeyError:
      userChangedPermissions.add(key)
  userDeniedPermissions = set()
  for key,value in oldDefaults.items():
    if (not key in userApproved):
      userDeniedPermissions.add(key)
  userSetPermissions = userChangedPermissions.union(userDeniedPermissions)
  addedOrChangedPermissions = {}
  for key,value in newDefaults.items():
    #                                                         added  or  changed
    if (not key in userSetPermissions) and ((not key in oldDefaults) or (key in oldDefaults and neq(oldDefaults[key],value))):
      addedOrChangedPermissions[key] = value
  droppedPermissions = []
  for key in oldDefaults.keys():
    if (key not in userSetPermissions) and (key not in newDefaults):
      droppedPermissions.append(key)
  return (droppedPermissions,addedOrChangedPermissions)

docs_header = """The permissions.json file format
================================

A permissions.json file is a file which describes the rights or permissions of a given subuser.  These permissions pertain mainly to that image's ability to interact with it's host operating system.

Each permissions.json file is to be a valid `json <https://www.ecma-international.org/publications/files/ECMA-ST/ECMA-404.pdf>`_ file containing a single json object.

All permissions are optional. If they are not included in the ``permissions.json`` file, than they are set to their default (and most restrictive) setting.

Setting a permission to an empty string is not a valid way of requesting it's default value.  If you want the default value, don't include the permission at all.

Permissions are categorized into 4 levels of permissiveness:

 1. conservative: These permissions should be 100% safe in all cases.
 2. moderate: These permissions have the potential to give the subuser access to some user data but not all.
 3. liberal: These permissions give the subuser access to some or all user data, and/or present a significantly increased risk of the subuser breaking out of the container.
 4. anarchistic: These permissions give the subuser full access to the system.

"""

depricatedPermissions = """
Depricated
----------

These permissions can still be used as top level permissions. However, they have now been rolled in as subpermissions to ``basic-common-permissions``.

 * ``stateful-home``: Changes that the subuser makes to it's home directory should be saved to a special subuser-homes directory.

  **Default**: ``false``

 * ``inherit-locale``: Automatically set the $LANG and $LANGUAGE environment variable in the container to the value outside of the container. Note: You do not have to set this if you have set ``basic-common-permissions``.

  **Default**: ``false``

 * ``inherit-timezone``: Automatically set the $TZ environment variable in the container to the value outside of the container.  Give the sub user read only access to the ``/etc/localtime`` file. Note: You do not have to set this if you have set ``basic-common-permissions``.

  **Default**: ``false``
"""

def get_default_description(default):
  default_descriptions = {
    "" : '""'
    ,True:"true"
    ,False:"false"
    ,None:"nil"
    ,OrderedDict:"{}"}
  try:
    return default_descriptions[default]
  except (KeyError,TypeError):
    try:
      return default_descriptions[type(default)]
    except KeyError:
      return str(default)

def get_permission_description(indent1,indent2,permission,permission_attrs):
  docs = ""
  try:
    docs += indent1+" ``"+permission+"``: "+permission_attrs["description"].replace("\n","\n"+indent2)+"\n"
  except KeyError:
    docs += indent1+" ``"+permission+"``:\n"
  docs += indent2+"Default: ``"+get_default_description(permission_attrs["default"])+"``\n\n"
  docs += indent2+"Types: ``"+"``, ``".join(getTypeDescriptions(permission_attrs["types"]))+"``\n\n"
  return docs


def getDocs():
  docs = docs_header
  for (level,level_description),permissions in supportedPermissions.items():
    docs += level_description+"\n"
    docs += "-"*len(level_description)+"\n"
    for permission,permission_attrs in permissions.items():
      docs += get_permission_description(" *","  ",permission,permission_attrs)
      if "subpermissions" in permission_attrs:
        docs += "  ``"+permission+"`` **subpermissions:**\n\n"
        for subpermission,subpermission_attrs in permission_attrs["subpermissions"].items():
          docs += get_permission_description("   -","    ",subpermission,subpermission_attrs)
  docs += depricatedPermissions
  return docs



# -*- coding: utf-8 -*-
# pylint: disable=no-init,old-style-class

"""
A file structure object is a read only view of a file structure, such as a real directory or a git tree.
"""

#external imports
import abc
import hashlib
import os
import stat
import sys
#internal imports
import subuserlib.test
import subuserlib.print
if subuserlib.test.testing:
  hashtestDir = subuserlib.test.hashtestDir

class FileStructure():
  __metaclass__ = abc.ABCMeta

  def ls(self, subfolder,objectType=None):
    """
    Returns a list of file and folder paths.
    Paths are relative to the root of the FileStructure.
    """
    self.assertLegalPath(subfolder)
    return self._ls(subfolder,objectType=objectType)

  @abc.abstractmethod
  def _ls(self,subfolder,objectType=None):
    pass

  def lsFiles(self,subfolder):
    """
    Returns a list of paths to files in the subfolder.
    Paths are relative to the root of the FileStructure.
    """
    self.assertLegalPath(subfolder)
    return self._lsFiles(subfolder)

  @abc.abstractmethod
  def _lsFiles(self,subfolder):
    pass

  def lsFolders(self,subfolder):
    """
    Returns a list of paths to folders in the subfolder.
    Paths are relative to the root of the FileStructure.
    """
    self.assertLegalPath(subfolder)
    return self._lsFolders(subfolder)

  @abc.abstractmethod
  def _lsFolders(self,subfolder):
    pass

  def exists(self,path):
    self.assertLegalPath(path)
    return self._exists(path)

  @abc.abstractmethod
  def _exists(self,path):
    pass

  def read(self,path):
    """
    Returns the contents of the given file.
    """
    self.assertLegalPath(path)
    return self._read(path)

  @abc.abstractmethod
  def _read(self,path):
    pass

  def readBinary(self,path):
    """
    Returns the contents of the given file.
    """
    self.assertLegalPath(path)
    return self._readBinary(path)

  @abc.abstractmethod
  def _readBinary(self,path):
    pass

  @abc.abstractmethod
  def getMode(self,path):
   self.assertLegalPath(path)
   return self._getMode(path)

  def getSize(self,path):
    self.assertLegalPath(path)
    return self._getSize(path)

  @abc.abstractmethod
  def _getSize(self,path):
    pass

  @abc.abstractmethod
  def isLegalSymlink(self,path):
    """
    If the path is a symlink. Make sure it doesn't point out of the structure.
    """
    pass

  def assertLegalPath(self,path):
    """
    Throw an exception if path is a relative path going up outside of the file structure.

    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)

    These should work.

    >>> fileStructure.assertLegalPath("./foo")
    >>> fileStructure.assertLegalPath("./bar/../foo")

    These shouldn't.

    >>> fileStructure.assertLegalPath("../foo")
    Traceback (most recent call last):
    ...
    OSError: ../foo does not exist in file structure.
    >>> fileStructure.assertLegalPath("./../foo")
    Traceback (most recent call last):
    ...
    OSError: ./../foo does not exist in file structure.
    >>> fileStructure.assertLegalPath("./bar/../../foo")
    Traceback (most recent call last):
    ...
    OSError: ./bar/../../foo does not exist in file structure.
    >>> fileStructure.assertLegalPath("./bar/../../../foo")
    Traceback (most recent call last):
    ...
    OSError: ./bar/../../../foo does not exist in file structure.

    Absolute paths are absolute nonos

    >>> fileStructure.assertLegalPath("/bar")
    Traceback (most recent call last):
    ...
    OSError: /bar does not exist in file structure.

    Checks for symlinks too.
    """
    if path.startswith("/") or os.path.relpath(path,"./").startswith(".."):
      raise IOError(path + " does not exist in file structure.")
    try:
      path = self.realpath(path)
      if not self.isLegalSymlink(path):
        raise IOError(path + " is a symlink which points outside of the filestructure which is not allowed.")
    except AttributeError:
      pass

  def getModeString(self,path):
    """
    Return the human readable mode string for the mode in octal notation.
    """
    self.assertLegalPath(path)
    octalMode = oct(self.getMode(path))
    if sys.version_info[0] == 2:
      octalMode = octalMode[1:]
    if sys.version_info[0] == 3:
      octalMode = octalMode[2:]
    return octalMode

  def hash(self,path,printDebugOutput=False):
    """
    Return the SHA512 hash of the file or directory.

    In the case of directectories, hashes alphabetically and hashes subdirectories first.

    Hashes the following:
      - Relative file path
      - File size as UTF-8 string
      - File contents

    NOTE: File mode is not included in hash because git doesn't actually support meaningful file modes. :(

    Return the hash as a hexidecimal string.

    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> fileStructure.hash("./")
    'b0cd63dd96b76d7a9c61e434b43f0eea408c2dd14dca1f436be0a56bf1f91aa75f4406b9fe9fb2025b84e3445f747a2680d56ca92f5b4fc28a98d8f70586cf15'
    """
    self.assertLegalPath(path)
    hashFunction = hashlib.sha512
    hash = hashFunction()
    # TODO - what about symlinks?
    # TODO - what about devices?
    # TODO - what about sockets?
    # TODO - what about named pipes?
    def hashFile(path):
      file_metadata_string = str(len(path))+" "+path+" "+str(self.getSize(path))+" "
      hash.update(file_metadata_string.encode("utf-8"))
      hash.update(self.readBinary(path))
      hash.update("\n".encode("utf-8"))
      if printDebugOutput:
        subuserlib.print.printWithoutCrashing(file_metadata_string.encode("utf-8").decode("utf-8")+self.readBinary(path).decode("utf-8","replace"))
    def hashDir(path):
      # Hash subdirectories
      subdirs = self.lsFolders(path)
      subdirs.sort()
      for subdir in subdirs:
        if subdir != ".git":
          hashDir(os.path.join(path,subdir))
      # Hash files
      files = self.lsFiles(path)
      files.sort()
      for fileToHash in files:
        hashFile(os.path.join(path,fileToHash))
    hashDir(path)
    return hash.hexdigest()

class BasicFileStructure(FileStructure):
  """
  A FileStructure backed by real files.
  """
  def __init__(self,path):
    self.path = path
    if not os.path.exists(path):
      raise FileNotFoundError(path+" does not exist.")

  def getPathInStructure(self,path):
    """
    Given a relative path within the file structure, return an absolute path.
    """
    return os.path.join(self.path,path)

  def _ls(self, subfolder,objectType=None):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(",".join(fileStructure.ls("./")))
    bar,blah
    """
    assert(objectType==None)
    paths = []
    path = self.getPathInStructure(subfolder)
    for path in os.listdir(path):
      paths.append(path)
    paths.sort()
    return paths

  def _lsFiles(self,subfolder):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(",".join(fileStructure.lsFiles("./")))
    blah
    >>> print(",".join(fileStructure.lsFiles("bar")))
    New York,abacus
    >>> print(",".join(fileStructure.lsFiles("./bar")))
    New York,abacus
    """
    files = []
    for path in self.ls(subfolder):
      if os.path.isfile(self.getPathInStructure(os.path.join(subfolder,path))):
        files.append(path)
    return files

  def _lsFolders(self,subfolder):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(",".join(fileStructure.lsFolders("./")))
    bar
    """
    folders = []
    for path in self.ls(subfolder):
      pathInStructure = self.getPathInStructure(os.path.join(subfolder,path))
      if os.path.isdir(pathInStructure):
        folders.append(path)
    return folders

  def _exists(self,path):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> fileStructure.exists("./blah")
    True
    >>> fileStructure.exists("./non-existant")
    False
    """
    return os.path.exists(self.getPathInStructure(path))

  def _read(self,path):
    """
    Reads in a file as a utf-8 string.

    >>> from subuserlib.classes.fileStructure import BasicFileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(fileStructure.read("./blah"))
    blahblah
    <BLANKLINE>
    
    You can only read in files that are actually inside the file structure.
    >>> fileStructure1 = BasicFileStructure(os.path.join(subuserlib.classes.fileStructure.hashtestDir,"bar"))
    >>> print(fileStructure1.read("../blah"))
    Traceback (most recent call last):
    ...
    OSError: ../blah does not exist in file structure.
    """
    with open(self.getPathInStructure(path),"r",encoding="utf-8") as fd:
      return fd.read()

  def _readBinary(self,path):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(fileStructure.read("./blah"))
    blahblah
    <BLANKLINE>
    """
    with open(self.getPathInStructure(path),"rb") as fd:
      return fd.read()

  def _getMode(self,path):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(fileStructure.getModeString("./blah"))
    100664
    """
    return os.stat(self.getPathInStructure(path))[stat.ST_MODE]

  def _getSize(self,path):
    """
    >>> from subuserlib.classes.fileStructure import FileStructure
    >>> import os
    >>> fileStructure = BasicFileStructure(subuserlib.classes.fileStructure.hashtestDir)
    >>> print(fileStructure.getSize("./blah"))
    9
    """
    return os.stat(self.getPathInStructure(path))[stat.ST_SIZE]

  def realpath(self,path):
    return os.path.realpath(self.getPathInStructure(path))

  def isLegalSymlink(self,path):
    return not os.path.relpath(path,self.path).startswith("..")
